"""
this module has the layer 2 config components ,
it has more business logic on the components like :
vlan.has_interface --> return true of that particular vlan has an interface
interface.is_svi --> return true of that particular interface is an svi interface
"""

from .base_component import FeatureSet, Feature


class Vlan(Feature):
    """single vlan model"""

    def __eq__(self, other):
        return isinstance(other, Vlan) and self.id == other.id

    @property
    def has_interface(self):
        if not self.interfaces:
            return False


class Vlans(FeatureSet):
    """
    device vlans manager class ,
    model attribute is to specify which model this class belong to
    """
    _feature_name = 'vlans'
    model = Vlan

    conf_template = "vlan.j2"



class Interface(Feature):
    """single interface model"""

    @property
    def is_svi(self):
        if self.hardware_type == 'EtherSVI':
            return True
        return False

    @property
    def is_port_channel(self):
        if self.hardware_type == 'EtherChannel':
            return True
        return False

    @property
    def is_loopback(self):
        if self.hardware_type == 'Loopback':
            return True
        return False

    @property
    def is_up(self):
        if self.link_status == 'up' and self.protocol_status == 'up':
            return True
        return False


class Interfaces(FeatureSet):
    """multiple interface models class"""
    _feature_name = 'interfaces'
    model = Interface
    conf_template = 'interface.j2'

    def bounce(self,id):
        self.cmds += [f"interface {id}" , "shutdown" , "no shutdown"]

    @property
    def down_list(self):
        return [i for i in self.all if not i.is_up]

    @property
    def svi_list(self):
        return [i for i in self.all if i.is_svi]

    @property
    def port_channel_list(self):
        return [i for i in self.all if i.is_port_channel]

    @property
    def loopback_list(self):
        return [i for i in self.all if i.is_loopback]


class CdpNeighbors(FeatureSet):
    _feature_name = 'cdp_neighbors'
    def get_cdps_for_port(self, port):
        return [i for i in self.all if i.local_port == port]

    def get_cdps_for_neighbor(self, neighbor_name):
        return [i for i in self.all if i.neighbor == neighbor_name]

    def get_cdp_by_ip(self, ip):
        return [i for i in self.all if i.neighbor_ip == ip]


class Vtp(Feature):
    pass


class Vpc(Feature):

    @property
    def is_up(self):
        if self.status == 'up':
            return True
        return False


class Vpcs(FeatureSet):
    model = Vpc
    _feature_name = 'vpcs'
    def get_vpc_by_port(self, port):
        return [i for i in self.all if i.port == port]

class Stp(Feature):
    pass

class Stps(FeatureSet):
    model = Stp
    _feature_name = 'spanning_tree'

    def get_by_vlan_id(self,id):
        return [stp for stp in self.all if stp.vlan_id == id]

    def get_by_interface(self,interface):
        for stp in self.all:
            if stp.interface == interface:
                return stp
    def get_by_port_status(self,status):
        """status = FWD|BLK"""
        return [stp for stp in self.all if stp.status == status]

    def get_blocked(self):
        return [stp for stp in self.all if stp.status == 'BLK']

    def get_forwarded(self):
        return [stp for stp in self.all if stp.status == 'FWD']

    def get_designated_ports(self):
        return [stp for stp in self.all if stp.status == 'Desg']

    def get_root_ports(self):
        return [stp for stp in self.all if stp.status == 'Root']









interface_oids = {
    'name': 'ifName',
    'mtu': 'ifMtu',
    'speed': 'ifSpeed',
    'mac': 'ifPhysAddress',
    'admin_status': 'ifAdminStatus',
    'oper_status': 'ifOperStatus',
    'description': 'ifAlias',
    'inErros': 'ifInErrors',
    'inDiscards': 'ifInDiscards',
    'outErrors': 'ifOutErrors',
    'outDiscards': 'ifOutDiscards'

}



from .layer2 import *
from .layer3 import *
from .system import *
from .security import *

__all__ = ['Vpcs', 'Vrfs', 'Vtp', 'Vlans', 'Interfaces', 'AccessLists', 'OspfNeighbors', 'BgpNeighbors', 'CdpNeighbors',
           'Modules']



from netmiko import ConnectHandler, NetmikoTimeoutError, NetMikoTimeoutException, \
    NetmikoAuthError, NetMikoAuthenticationException


class CiscoSDKSSHAuthenticationError(Exception):
    pass


class CiscoSDKSSHTimeoutError(Exception):
    pass


def _netmiko_connect(connection_dict):
    """
    to handel netmiko connection
    :param connection_dict: netmiko connection dict
    :return: netmiko connection object
    """
    try:
        connection = ConnectHandler(**connection_dict)

    except (NetMikoTimeoutException, NetmikoTimeoutError) as e:
        print(f"Time out while connecting to device {connection_dict['ip']}")
        raise CiscoSDKSSHTimeoutError(f"ERROR : Time out while connecting to device {connection_dict['ip']}")
    except (NetmikoAuthError, NetMikoAuthenticationException) as e:
        print(f"ERROR : ssh authentication failed to device {connection_dict['ip']}")
        raise CiscoSDKSSHAuthenticationError(f"SSH authentication failed to device {connection_dict['ip']}")
    return connection


class SSHManager:
    """
    SSH Context Manager
    """

    def __init__(self, connection_dict):

        self.conn_dict = connection_dict
        self.conn = self.connect()

    def __enter__(self):
        self.conn = self.connect()
        if not self.conn:
            return False
        return self

    def __exit__(self, *args):
        if not self.conn:
            return
        self.conn.disconnect()

    def connect(self):
        print(f"connecting to {self.conn_dict['ip']}")
        return _netmiko_connect(connection_dict=self.conn_dict)

    def save_config(self):
        self.conn.save_config()

    def send_commands_list(self, cmd):

        output = self.conn.send_config_set(cmd, strip_command=True)

        return output

    def get_command(self, cmd):
        try:
            output = self.conn.send_command(cmd, use_textfsm=True)
        except ValueError:
            raise ValueError(
                "invalid ~/ntc-templates/templates , please make sure to download download ntc-templates/templates from "
                "https://github.com/Ali-aqrabawi/cisco_sdk/templates and add templates dir to ~/ntc-templates/templates ")
        return output



name = "cscmiko"
import sys
from .devices import *

__version__ = '0.0.2'
# Verify Python Version
try:
    if not sys.version_info.major == 3:
        raise RuntimeError('cscmiko requires Python3')
except AttributeError:
    raise RuntimeError('cscmiko requires Python3')



"""
this module has the system config components like cpu
it has more business logic on the components
"""

from .base_component import Feature, FeatureSet


class Cpu(Feature):
    @property
    def is_high(self):
        if int(self.cpu_5_sec) > 80:
            return True
        return False

    def is_higher_than(self, utilization):
        if int(self.cpu_5_sec) > utilization:
            return True
        return False


class Module(Feature):

    @property
    def is_active(self):
        if self.status == 'active':
            return True
        return False

    @property
    def is_ok(self):
        if self.status == 'ok':
            return True
        return False

    @property
    def is_down(self):
        if self.status == 'down':
            return True
        return False


class Modules(FeatureSet):
    """
    cisco "show module" for nexus and 6k
    """
    _feature_name = 'modules'
    model = Module

    def get_module_by_model(self, model):
        for module in self.all:
            if model in module.model:
                return module

    @property
    def down_list(self):
        return [i for i in self.all if i.is_down]



import re
import os
from jinja2 import Environment, FileSystemLoader

def clean(x):
    def remove(x):
        string = x.strip(" ")
        return string
    return list(filter(None,map(remove,x)))

def check_config_result(output):
    """
    take the output result of executed commands on the device and check if any commands has failed
    :param output: str: result of executed commands
    :return: tuple: Bool,str: True if no errors found , str is the err_msg
    """
    print("Validating config execution")
    err_lines = []
    output = output.strip('config term')
    output = output.strip('Enter configuration commands, one per line.  End with CNTL/Z.')
    output_lines = output.splitlines()
    # Remove the empty lines
    output_lines = list(filter(None, output_lines))

    # Extract the Errors and warnings
    for line in output_lines:
        f = re.match(".*#", line)
        if not f:
            err_lines.append(line)

    # Classify errors and warnnings
    for err in err_lines:

        if "^" in err:
            print("Config failed , reason = syntax error ")

            return False, "internal error, (wrong syntax)"

        if 'ERROR' in err:
            print(f"config failed, reason = {err}")
            return False, err



    return True, err_lines


def render_command(template_name, variables):
    """
    render commands and return list of rendered cmds

    :param template_name: jinja template name
    :param variables: command variables
    :return: cmd: list of rendered commands
    """
    # get cisco_sdk/ path
    PROJECT_DIR = os.path.dirname(os.path.dirname(__file__))
    TEMPLATES_DIR = os.path.join(PROJECT_DIR, "features/configuration_templates")
    # render the jinja2 template from a string
    env = Environment(loader=FileSystemLoader(searchpath=TEMPLATES_DIR))

    template = env.get_template(template_name)

    cmd = template.render(**variables)
    cmd = cmd.split('\n')


    return clean(cmd)
